@@ -17,7 +17,7 @@
 %                               September 2014                                %
 %                                                                             %
 %                                                                             %
-%  Copyright 1999-2021 ImageMagick Studio LLC, a non-profit organization      %
+%  Copyright @ 1999 ImageMagick Studio LLC, a non-profit organization         %
 %  dedicated to making software imaging solutions freely available.           %
 %                                                                             %
 %  You may not use this file except in compliance with the License.  You may  %
@@ -124,6 +124,26 @@ static int CCObjectInfoCompare(const void *x,const void *y)
 
   p=(CCObjectInfo *) x;
   q=(CCObjectInfo *) y;
+  if (p->key == -5)
+    return((int) (q->bounding_box.y-(ssize_t) p->bounding_box.y));
+  if (p->key == -4)
+    return((int) (q->bounding_box.x-(ssize_t) p->bounding_box.x));
+  if (p->key == -3)
+    return((int) (q->bounding_box.height-p->bounding_box.height));
+  if (p->key == -2)
+    return((int) (q->bounding_box.width-p->bounding_box.width));
+  if (p->key == -1)
+    return((int) (q->area-(ssize_t) p->area));
+  if (p->key == 1)
+    return((int) (p->area-(ssize_t) q->area));
+  if (p->key == 2)
+    return((int) (p->bounding_box.width-q->bounding_box.width));
+  if (p->key == 3)
+    return((int) (p->bounding_box.height-q->bounding_box.height));
+  if (p->key == 4)
+    return((int) (p->bounding_box.x-(ssize_t) q->bounding_box.x));
+  if (p->key == 5)
+    return((int) (p->bounding_box.y-(ssize_t) q->bounding_box.y));
   return((int) (q->area-(ssize_t) p->area));
 }
 
@@ -162,7 +182,7 @@ static void PerimeterThreshold(const Image *component_image,
       continue;
     component_view=AcquireAuthenticCacheView(component_image,exception);
     bounding_box=object[i].bounding_box;
-    for (y=(-1); y < (ssize_t) bounding_box.height+1; y++)
+    for (y=(-1); y < (ssize_t) bounding_box.height; y++)
     {
       const Quantum
         *magick_restrict p;
@@ -170,8 +190,6 @@ static void PerimeterThreshold(const Image *component_image,
       ssize_t
         x;
 
-      if (status == MagickFalse)
-        continue;
       p=GetCacheViewVirtualPixels(component_view,bounding_box.x-1,
         bounding_box.y+y,bounding_box.width+2,2,exception);
       if (p == (const Quantum *) NULL)
@@ -179,17 +197,17 @@ static void PerimeterThreshold(const Image *component_image,
           status=MagickFalse;
           break;
         }
-      for (x=(-1); x < (ssize_t) bounding_box.width+1; x++)
+      for (x=(-1); x < (ssize_t) bounding_box.width; x++)
       {
         Quantum
           pixels[4];
 
-        ssize_t
-          v;
-
         size_t
           foreground;
 
+        ssize_t
+          v;
+
         /*
           An Algorithm for Calculating Objects’ Shape Features in Binary
           Images, Lifeng He, Yuyan Chao.
@@ -205,9 +223,9 @@ static void PerimeterThreshold(const Image *component_image,
             ssize_t
               offset;
 
-            offset=v*(bounding_box.width+2)*
-              GetPixelChannels(component_image)+u*
-              GetPixelChannels(component_image);
+            offset=v*((ssize_t) bounding_box.width+2)*
+              (ssize_t) GetPixelChannels(component_image)+u*
+              (ssize_t) GetPixelChannels(component_image);
             pixels[2*v+u]=GetPixelIndex(component_image,p+offset);
             if ((ssize_t) pixels[2*v+u] == i)
               foreground++;
@@ -218,10 +236,8 @@ static void PerimeterThreshold(const Image *component_image,
         else
           if (foreground == 2)
             {
-              if ((((ssize_t) pixels[0] == i) &&
-                    ((ssize_t) pixels[3] == i)) ||
-                  (((ssize_t) pixels[1] == i) &&
-                    ((ssize_t) pixels[2] == i)))
+              if ((((ssize_t) pixels[0] == i) && ((ssize_t) pixels[3] == i)) ||
+                  (((ssize_t) pixels[1] == i) && ((ssize_t) pixels[2] == i)))
                 pattern[0]++;  /* diagonal */
               else
                 pattern[2]++;
@@ -229,7 +245,7 @@ static void PerimeterThreshold(const Image *component_image,
           else
             if (foreground == 3)
               pattern[3]++;
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
     component_view=DestroyCacheView(component_view);
@@ -281,8 +297,6 @@ static void CircularityThreshold(const Image *component_image,
       ssize_t
         x;
 
-      if (status == MagickFalse)
-        continue;
       p=GetCacheViewVirtualPixels(component_view,bounding_box.x-1,
         bounding_box.y+y,bounding_box.width+2,2,exception);
       if (p == (const Quantum *) NULL)
@@ -316,9 +330,9 @@ static void CircularityThreshold(const Image *component_image,
             ssize_t
               offset;
 
-            offset=v*(bounding_box.width+2)*
-              GetPixelChannels(component_image)+u*
-              GetPixelChannels(component_image);
+            offset=v*((ssize_t) bounding_box.width+2)*
+              (ssize_t) GetPixelChannels(component_image)+u*
+              (ssize_t) GetPixelChannels(component_image);
             pixels[2*v+u]=GetPixelIndex(component_image,p+offset);
             if ((ssize_t) pixels[2*v+u] == i)
               foreground++;
@@ -329,10 +343,8 @@ static void CircularityThreshold(const Image *component_image,
         else
           if (foreground == 2)
             {
-              if ((((ssize_t) pixels[0] == i) &&
-                    ((ssize_t) pixels[3] == i)) ||
-                  (((ssize_t) pixels[1] == i) &&
-                    ((ssize_t) pixels[2] == i)))
+              if ((((ssize_t) pixels[0] == i) && ((ssize_t) pixels[3] == i)) ||
+                  (((ssize_t) pixels[1] == i) && ((ssize_t) pixels[2] == i)))
                 pattern[0]++;  /* diagonal */
               else
                 pattern[2]++;
@@ -340,7 +352,7 @@ static void CircularityThreshold(const Image *component_image,
           else
             if (foreground == 3)
               pattern[3]++;
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
     component_view=DestroyCacheView(component_view);
@@ -402,8 +414,6 @@ static void MajorAxisThreshold(const Image *component_image,
     bounding_box=object[i].bounding_box;
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
-      if (status == MagickFalse)
-        continue;
       p=GetCacheViewVirtualPixels(component_view,bounding_box.x,
         bounding_box.y+y,bounding_box.width,1,exception);
       if (p == (const Quantum *) NULL)
@@ -419,11 +429,11 @@ static void MajorAxisThreshold(const Image *component_image,
             M10+=x;
             M01+=y;
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
-    centroid.x=M10*PerceptibleReciprocal(M00);
-    centroid.y=M01*PerceptibleReciprocal(M00);
+    centroid.x=M10*MagickSafeReciprocal(M00);
+    centroid.y=M01*MagickSafeReciprocal(M00);
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
       if (status == MagickFalse)
@@ -443,12 +453,12 @@ static void MajorAxisThreshold(const Image *component_image,
             M20+=(x-centroid.x)*(x-centroid.x);
             M02+=(y-centroid.y)*(y-centroid.y);
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
     component_view=DestroyCacheView(component_view);
-    object[i].metric[metric_index]=sqrt((2.0*PerceptibleReciprocal(M00))*((M20+M02)+
-      sqrt(4.0*M11*M11+(M20-M02)*(M20-M02))));
+    object[i].metric[metric_index]=sqrt((2.0*MagickSafeReciprocal(M00))*
+      ((M20+M02)+sqrt(4.0*M11*M11+(M20-M02)*(M20-M02))));
   }
 }
 
@@ -503,8 +513,6 @@ static void MinorAxisThreshold(const Image *component_image,
     bounding_box=object[i].bounding_box;
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
-      if (status == MagickFalse)
-        continue;
       p=GetCacheViewVirtualPixels(component_view,bounding_box.x,
         bounding_box.y+y,bounding_box.width,1,exception);
       if (p == (const Quantum *) NULL)
@@ -520,11 +528,11 @@ static void MinorAxisThreshold(const Image *component_image,
             M10+=x;
             M01+=y;
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
-    centroid.x=M10*PerceptibleReciprocal(M00);
-    centroid.y=M01*PerceptibleReciprocal(M00);
+    centroid.x=M10*MagickSafeReciprocal(M00);
+    centroid.y=M01*MagickSafeReciprocal(M00);
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
       if (status == MagickFalse)
@@ -544,12 +552,12 @@ static void MinorAxisThreshold(const Image *component_image,
             M20+=(x-centroid.x)*(x-centroid.x);
             M02+=(y-centroid.y)*(y-centroid.y);
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
     component_view=DestroyCacheView(component_view);
-    object[i].metric[metric_index]=sqrt((2.0*PerceptibleReciprocal(M00))*((M20+M02)-
-      sqrt(4.0*M11*M11+(M20-M02)*(M20-M02))));
+    object[i].metric[metric_index]=sqrt((2.0*MagickSafeReciprocal(M00))*
+      ((M20+M02)-sqrt(4.0*M11*M11+(M20-M02)*(M20-M02))));
   }
 }
 
@@ -605,8 +613,6 @@ static void EccentricityThreshold(const Image *component_image,
     bounding_box=object[i].bounding_box;
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
-      if (status == MagickFalse)
-        continue;
       p=GetCacheViewVirtualPixels(component_view,bounding_box.x,
         bounding_box.y+y,bounding_box.width,1,exception);
       if (p == (const Quantum *) NULL)
@@ -622,11 +628,11 @@ static void EccentricityThreshold(const Image *component_image,
             M10+=x;
             M01+=y;
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
-    centroid.x=M10*PerceptibleReciprocal(M00);
-    centroid.y=M01*PerceptibleReciprocal(M00);
+    centroid.x=M10*MagickSafeReciprocal(M00);
+    centroid.y=M01*MagickSafeReciprocal(M00);
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
       if (status == MagickFalse)
@@ -646,16 +652,16 @@ static void EccentricityThreshold(const Image *component_image,
             M20+=(x-centroid.x)*(x-centroid.x);
             M02+=(y-centroid.y)*(y-centroid.y);
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
     component_view=DestroyCacheView(component_view);
-    ellipse_axis.x=sqrt((2.0*PerceptibleReciprocal(M00))*((M20+M02)+
+    ellipse_axis.x=sqrt((2.0*MagickSafeReciprocal(M00))*((M20+M02)+
       sqrt(4.0*M11*M11+(M20-M02)*(M20-M02))));
-    ellipse_axis.y=sqrt((2.0*PerceptibleReciprocal(M00))*((M20+M02)-
+    ellipse_axis.y=sqrt((2.0*MagickSafeReciprocal(M00))*((M20+M02)-
       sqrt(4.0*M11*M11+(M20-M02)*(M20-M02))));
     object[i].metric[metric_index]=sqrt(1.0-(ellipse_axis.y*ellipse_axis.y*
-      PerceptibleReciprocal(ellipse_axis.x*ellipse_axis.x)));
+      MagickSafeReciprocal(ellipse_axis.x*ellipse_axis.x)));
   }
 }
 
@@ -710,8 +716,6 @@ static void AngleThreshold(const Image *component_image,
     bounding_box=object[i].bounding_box;
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
-      if (status == MagickFalse)
-        continue;
       p=GetCacheViewVirtualPixels(component_view,bounding_box.x,
         bounding_box.y+y,bounding_box.width,1,exception);
       if (p == (const Quantum *) NULL)
@@ -727,11 +731,11 @@ static void AngleThreshold(const Image *component_image,
             M10+=x;
             M01+=y;
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
-    centroid.x=M10*PerceptibleReciprocal(M00);
-    centroid.y=M01*PerceptibleReciprocal(M00);
+    centroid.x=M10*MagickSafeReciprocal(M00);
+    centroid.y=M01*MagickSafeReciprocal(M00);
     for (y=0; y < (ssize_t) bounding_box.height; y++)
     {
       if (status == MagickFalse)
@@ -751,12 +755,12 @@ static void AngleThreshold(const Image *component_image,
             M20+=(x-centroid.x)*(x-centroid.x);
             M02+=(y-centroid.y)*(y-centroid.y);
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
     component_view=DestroyCacheView(component_view);
     object[i].metric[metric_index]=RadiansToDegrees(1.0/2.0*atan(2.0*M11*
-      PerceptibleReciprocal(M20-M02)));
+      MagickSafeReciprocal(M20-M02)));
     if (fabs(M11) < 0.0)
         {
           if ((fabs(M20-M02) >= 0.0) && ((M20-M02) < 0.0))
@@ -815,9 +819,6 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
   MatrixInfo
     *equivalences;
 
-  ssize_t
-    i;
-
   size_t
     size;
 
@@ -828,6 +829,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
     dx,
     dy,
     first,
+    i,
     last,
     n,
     step,
@@ -838,10 +840,10 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
   */
   assert(image != (Image *) NULL);
   assert(image->signature == MagickCoreSignature);
-  if (image->debug != MagickFalse)
-    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickCoreSignature);
+  if (IsEventLogging() != MagickFalse)
+    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   if (objects != (CCObjectInfo **) NULL)
     *objects=(CCObjectInfo *) NULL;
   component_image=CloneImage(image,0,0,MagickTrue,exception);
@@ -913,7 +915,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
           status=MagickFalse;
           continue;
         }
-      p+=GetPixelChannels(image)*image->columns;
+      p+=(ptrdiff_t) GetPixelChannels(image)*image->columns;
       for (x=0; x < (ssize_t) image->columns; x++)
       {
         PixelInfo
@@ -935,22 +937,22 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
         if (((x+dx) < 0) || ((x+dx) >= (ssize_t) image->columns) ||
             ((y+dy) < 0) || ((y+dy) >= (ssize_t) image->rows))
           {
-            p+=GetPixelChannels(image);
+            p+=(ptrdiff_t) GetPixelChannels(image);
             continue;
           }
-        neighbor_offset=dy*(GetPixelChannels(image)*image->columns)+dx*
-          GetPixelChannels(image);
+        neighbor_offset=dy*((ssize_t) GetPixelChannels(image)*(ssize_t)
+          image->columns)+dx*(ssize_t) GetPixelChannels(image);
         GetPixelInfoPixel(image,p+neighbor_offset,&target);
         if (IsFuzzyEquivalencePixelInfo(&pixel,&target) == MagickFalse)
           {
-            p+=GetPixelChannels(image);
+            p+=(ptrdiff_t) GetPixelChannels(image);
             continue;
           }
         /*
           Resolve this equivalence.
         */
-        offset=y*image->columns+x;
-        neighbor_offset=dy*image->columns+dx;
+        offset=y*(ssize_t) image->columns+x;
+        neighbor_offset=dy*(ssize_t) image->columns+dx;
         ox=offset;
         status=GetMatrixElement(equivalences,ox,0,&obj);
         while (obj != ox)
@@ -989,8 +991,9 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
           status=GetMatrixElement(equivalences,oy,0,&obj);
           status=SetMatrixElement(equivalences,oy,0,&root);
         }
-        status=SetMatrixElement(equivalences,y*image->columns+x,0,&root);
-        p+=GetPixelChannels(image);
+        status=SetMatrixElement(equivalences,y*(ssize_t) image->columns+x,0,
+          &root);
+        p+=(ptrdiff_t) GetPixelChannels(image);
       }
     }
   }
@@ -1026,7 +1029,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
         id,
         offset;
 
-      offset=y*image->columns+x;
+      offset=y*(ssize_t) image->columns+x;
       status=GetMatrixElement(equivalences,offset,0,&id);
       if (id != offset)
         status=GetMatrixElement(equivalences,id,0,&id);
@@ -1045,19 +1048,19 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
         object[id].bounding_box.y=y;
       if (y >= (ssize_t) object[id].bounding_box.height)
         object[id].bounding_box.height=(size_t) y;
-      object[id].color.red+=QuantumScale*GetPixelRed(image,p);
-      object[id].color.green+=QuantumScale*GetPixelGreen(image,p);
-      object[id].color.blue+=QuantumScale*GetPixelBlue(image,p);
+      object[id].color.red+=QuantumScale*(double) GetPixelRed(image,p);
+      object[id].color.green+=QuantumScale*(double) GetPixelGreen(image,p);
+      object[id].color.blue+=QuantumScale*(double) GetPixelBlue(image,p);
       if (image->alpha_trait != UndefinedPixelTrait)
-        object[id].color.alpha+=QuantumScale*GetPixelAlpha(image,p);
+        object[id].color.alpha+=QuantumScale*(double) GetPixelAlpha(image,p);
       if (image->colorspace == CMYKColorspace)
-        object[id].color.black+=QuantumScale*GetPixelBlack(image,p);
+        object[id].color.black+=QuantumScale*(double) GetPixelBlack(image,p);
       object[id].centroid.x+=x;
       object[id].centroid.y+=y;
       object[id].area++;
       SetPixelIndex(component_image,(Quantum) id,q);
-      p+=GetPixelChannels(image);
-      q+=GetPixelChannels(component_image);
+      p+=(ptrdiff_t) GetPixelChannels(image);
+      q+=(ptrdiff_t) GetPixelChannels(component_image);
     }
     if (n > (ssize_t) MaxColormapSize)
       break;
@@ -1090,8 +1093,10 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
   component_image->colors=(size_t) n;
   for (i=0; i < (ssize_t) component_image->colors; i++)
   {
-    object[i].bounding_box.width-=(object[i].bounding_box.x-1);
-    object[i].bounding_box.height-=(object[i].bounding_box.y-1);
+    object[i].bounding_box.width=(size_t) ((ssize_t)
+      object[i].bounding_box.width-(object[i].bounding_box.x-1));
+    object[i].bounding_box.height=(size_t) ((ssize_t)
+      object[i].bounding_box.height-(object[i].bounding_box.y-1));
     object[i].color.red/=(QuantumScale*object[i].area);
     object[i].color.green/=(QuantumScale*object[i].area);
     object[i].color.blue/=(QuantumScale*object[i].area);
@@ -1116,7 +1121,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max area threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
       for (i=0; i < (ssize_t) component_image->colors; i++)
         if (((object[i].area < min_threshold) ||
              (object[i].area >= max_threshold)) && (i != background_id))
@@ -1284,9 +1289,9 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max perimeter threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
       metrics[++n]="perimeter";
-      PerimeterThreshold(image,object,n,exception);
+      PerimeterThreshold(component_image,object,n,exception);
       for (i=0; i < (ssize_t) component_image->colors; i++)
         if (((object[i].metric[n] < min_threshold) ||
              (object[i].metric[n] >= max_threshold)) && (i != background_id))
@@ -1298,12 +1303,12 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max circularity threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
       metrics[++n]="circularity";
-      CircularityThreshold(image,object,n,exception);
+      CircularityThreshold(component_image,object,n,exception);
       for (i=0; i < (ssize_t) component_image->colors; i++)
         if (((object[i].metric[n] < min_threshold) ||
-              (object[i].metric[n] >= max_threshold)) && (i != background_id))
+             (object[i].metric[n] >= max_threshold)) && (i != background_id))
           object[i].merge=MagickTrue;
     }
   artifact=GetImageArtifact(image,"connected-components:diameter-threshold");
@@ -1312,13 +1317,13 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max diameter threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
       metrics[++n]="diameter";
       for (i=0; i < (ssize_t) component_image->colors; i++)
       {
         object[i].metric[n]=ceil(sqrt(4.0*object[i].area/MagickPI)-0.5);
         if (((object[i].metric[n] < min_threshold) ||
-             (object[i].metric[n] >= max_threshold)) && (i != background_id))
+            (object[i].metric[n] >= max_threshold)) && (i != background_id))
           object[i].merge=MagickTrue;
       }
     }
@@ -1328,12 +1333,12 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max ellipse major threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
       metrics[++n]="major-axis";
       MajorAxisThreshold(component_image,object,n,exception);
       for (i=0; i < (ssize_t) component_image->colors; i++)
         if (((object[i].metric[n] < min_threshold) ||
-              (object[i].metric[n] >= max_threshold)) && (i != background_id))
+             (object[i].metric[n] >= max_threshold)) && (i != background_id))
           object[i].merge=MagickTrue;
     }
   artifact=GetImageArtifact(image,"connected-components:minor-axis-threshold");
@@ -1342,12 +1347,12 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max ellipse minor threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
       metrics[++n]="minor-axis";
       MinorAxisThreshold(component_image,object,n,exception);
       for (i=0; i < (ssize_t) component_image->colors; i++)
         if (((object[i].metric[n] < min_threshold) ||
-              (object[i].metric[n] >= max_threshold)) && (i != background_id))
+             (object[i].metric[n] >= max_threshold)) && (i != background_id))
           object[i].merge=MagickTrue;
     }
   artifact=GetImageArtifact(image,"connected-components:eccentricity-threshold");
@@ -1356,8 +1361,8 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max eccentricity threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
-      metrics[++n]="eccentricy";
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      metrics[++n]="eccentricity";
       EccentricityThreshold(component_image,object,n,exception);
       for (i=0; i < (ssize_t) component_image->colors; i++)
         if (((object[i].metric[n] < min_threshold) ||
@@ -1370,7 +1375,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       /*
         Merge any object not within the min and max ellipse angle threshold.
       */
-      (void) sscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
+      (void) MagickSscanf(artifact,"%lf%*[ -]%lf",&min_threshold,&max_threshold);
       metrics[++n]="angle";
       AngleThreshold(component_image,object,n,exception);
       for (i=0; i < (ssize_t) component_image->colors; i++)
@@ -1383,6 +1388,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
   */
   component_view=AcquireAuthenticCacheView(component_image,exception);
   object_view=AcquireVirtualCacheView(component_image,exception);
+  (void) SetCacheViewVirtualPixelMethod(object_view,TileVirtualPixelMethod);
   for (i=0; i < (ssize_t) component_image->colors; i++)
   {
     RectangleInfo
@@ -1423,7 +1429,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
         }
       for (x=0; x < (ssize_t) bounding_box.width; x++)
       {
-        size_t
+        ssize_t
           k;
 
         if (status == MagickFalse)
@@ -1453,7 +1459,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
             if (j != i)
               object[j].census++;
           }
-        p+=GetPixelChannels(component_image);
+        p+=(ptrdiff_t) GetPixelChannels(component_image);
       }
     }
     /*
@@ -1485,7 +1491,7 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
       {
         if ((ssize_t) GetPixelIndex(component_image,q) == i)
           SetPixelIndex(component_image,(Quantum) id,q);
-        q+=GetPixelChannels(component_image);
+        q+=(ptrdiff_t) GetPixelChannels(component_image);
       }
       if (SyncCacheViewAuthenticPixels(component_view,exception) == MagickFalse)
         status=MagickFalse;
@@ -1507,6 +1513,10 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
   if ((IsStringTrue(artifact) != MagickFalse) ||
       (objects != (CCObjectInfo **) NULL))
     {
+      ssize_t
+        key,
+        order;
+
       /*
         Report statistics on each unique object.
       */
@@ -1556,17 +1566,41 @@ MagickExport Image *ConnectedComponentsImage(const Image *image,
           object[id].centroid.x+=x;
           object[id].centroid.y+=y;
           object[id].area++;
-          p+=GetPixelChannels(component_image);
+          p+=(ptrdiff_t) GetPixelChannels(component_image);
         }
       }
       for (i=0; i < (ssize_t) component_image->colors; i++)
       {
-        object[i].bounding_box.width-=(object[i].bounding_box.x-1);
-        object[i].bounding_box.height-=(object[i].bounding_box.y-1);
+        object[i].bounding_box.width=(size_t) ((ssize_t)
+          object[i].bounding_box.width-(object[i].bounding_box.x-1));
+        object[i].bounding_box.height=(size_t) ((ssize_t)
+          object[i].bounding_box.height-(object[i].bounding_box.y-1));
         object[i].centroid.x=object[i].centroid.x/object[i].area;
         object[i].centroid.y=object[i].centroid.y/object[i].area;
       }
       component_view=DestroyCacheView(component_view);
+      order=1;
+      artifact=GetImageArtifact(image,"connected-components:sort-order");
+      if (artifact != (const char *) NULL)
+        if (LocaleCompare(artifact,"decreasing") == 0)
+          order=(-1);
+      key=0;
+      artifact=GetImageArtifact(image,"connected-components:sort");
+      if (artifact != (const char *) NULL)
+        {
+          if (LocaleCompare(artifact,"area") == 0)
+            key=1;
+          if (LocaleCompare(artifact,"width") == 0)
+            key=2;
+          if (LocaleCompare(artifact,"height") == 0)
+            key=3;
+          if (LocaleCompare(artifact,"x") == 0)
+            key=4;
+          if (LocaleCompare(artifact,"y") == 0)
+            key=5;
+        }
+      for (i=0; i < (ssize_t) component_image->colors; i++)
+         object[i].key=order*key;
       qsort((void *) object,component_image->colors,sizeof(*object),
         CCObjectInfoCompare);
       if (objects == (CCObjectInfo **) NULL)
@@ -1670,10 +1704,10 @@ MagickExport Image *IntegralImage(const Image *image,ExceptionInfo *exception)
   */
   assert(image != (const Image *) NULL);
   assert(image->signature == MagickCoreSignature);
-  if (image->debug != MagickFalse)
-    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   assert(exception != (ExceptionInfo *) NULL);
   assert(exception->signature == MagickCoreSignature);
+  if (IsEventLogging() != MagickFalse)
+    (void) LogMagickEvent(TraceEvent,GetMagickModule(),"%s",image->filename);
   integral_image=CloneImage(image,0,0,MagickTrue,exception);
   if (integral_image == (Image *) NULL)
     return((Image *) NULL);
@@ -1732,15 +1766,15 @@ MagickExport Image *IntegralImage(const Image *image,ExceptionInfo *exception)
           continue;
         sum=(double) q[i];
         if (x > 0)
-          sum+=(q-GetPixelChannels(integral_image))[i];
+          sum+=(double) (q-GetPixelChannels(integral_image))[i];
         if (y > 0)
-          sum+=p[i];
+          sum+=(double) p[i];
         if ((x > 0) && (y > 0))
-          sum-=(p-GetPixelChannels(integral_image))[i];
+          sum-=(double) (p-GetPixelChannels(integral_image))[i];
         q[i]=ClampToQuantum(sum);
       }
-      p+=GetPixelChannels(integral_image);
-      q+=GetPixelChannels(integral_image);
+      p+=(ptrdiff_t) GetPixelChannels(integral_image);
+      q+=(ptrdiff_t) GetPixelChannels(integral_image);
     }
     sync=SyncCacheViewAuthenticPixels(integral_view,exception);
     if (sync == MagickFalse)
